// PerformanceMonster
// Copyright (C) 2005  Ole Friis <olefriis@gmail.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

package performancemonster.loghandler;

import java.net.InetAddress;
import java.net.Socket;
import java.io.IOException;
import java.io.ObjectOutputStream;

import org.apache.log4j.helpers.LogLog;
import org.apache.log4j.spi.LoggingEvent;
import org.apache.log4j.AppenderSkeleton;

public class MonsterAppender extends AppenderSkeleton
{
  private static final int DEFAULT_PORT = 4560;
  private static final int DEFAULT_RECONNECTION_DELAY = 30000;

  private InetAddress address;
  private int port = DEFAULT_PORT;
  private ObjectOutputStream oos;
  private int reconnectionDelay = DEFAULT_RECONNECTION_DELAY;
  private Connector connector;
  private int counter = 0;

  // reset the ObjectOutputStream every 70 calls
  // private static final int RESET_FREQUENCY = 70;
  private static final int RESET_FREQUENCY = 1;

  public MonsterAppender()
  {
    // Intentionally left blank
  }

  /**
   * The <b>RemoteHost</b> option takes a string value which should be the host
   * name of the server where a
   * <code>link org.apache.log4j.net.SocketNode</code> is running.
   */
  public void setRemoteHost(String host)
  {
    address = getAddressByName(host);
  }

  /**
   * The <b>Port</b> option takes a positive integer representing the port
   * where the server is waiting for connections.
   */
  public void setPort(int port)
  {
    this.port = port;
  }

  /**
   * The <b>ReconnectionDelay</b> option takes a positive integer representing
   * the number of milliseconds to wait between each failed connection attempt
   * to the server. The default value of this option is 30000 which corresponds
   * to 30 seconds.
   * <p>
   * Setting this option to zero turns off reconnection capability.
   */
  public void setReconnectionDelay(int delay)
  {
    this.reconnectionDelay = delay;
  }

  /**
   * Connect to the specified <b>RemoteHost</b> and <b>Port</b>.
   */
  public void activateOptions()
  {
    connect();
  }

  /**
   * Close this appender.
   * <p>
   * This will mark the appender as closed and call then {@link #cleanUp}
   * method.
   */
  synchronized public void close()
  {
    if (closed)
      return;

    this.closed = true;
    cleanUp();
  }

  /**
   * MonsterAppender does not use a layout. Hence, this method returns
   * <code>false</code>.
   */
  public boolean requiresLayout()
  {
    return false;
  }

  public void append(LoggingEvent event)
  {
    if (event == null)
      return;

    LogMessage logMessage = new LogMessage(event.getMessage().toString());
    sendLogMessage(logMessage);
  }

  public void sendLogMessage(LogMessage object)
  {
    System.out.println("Sending message to " + address + ", oos: " + oos);
    if (address == null) {
      errorHandler.error("No remote host is set for SocketAppender named \""
          + this.name + "\".");
      return;
    }

    if (oos != null) {
      try {
        oos.writeObject(object);
        oos.flush();
        if (++counter >= RESET_FREQUENCY) {
          counter = 0;
          // Failing to reset the object output stream every now and
          // then creates a serious memory leak.
          oos.reset();
        }
      } catch (IOException e) {
        oos = null;
        LogLog.warn("Detected problem with connection: " + e);
        if (reconnectionDelay > 0) {
          fireConnector();
        }
      }
    }
  }

  /**
   * Drop the connection to the remote host and release the underlying connector
   * thread if it has been created
   */
  private void cleanUp()
  {
    if (oos != null) {
      try {
        oos.close();
      } catch (IOException e) {
        LogLog.error("Could not close oos.", e);
      }
      oos = null;
    }
    if (connector != null) {
      connector.interrupted = true;
      connector = null; // allow gc
    }
  }

  private void connect()
  {
    if (this.address == null)
      return;
    try {
      // First, close the previous connection if any.
      cleanUp();
      oos = new ObjectOutputStream(new Socket(address, port).getOutputStream());
    } catch (IOException e) {

      String msg = "Could not connect to remote log4j server at ["
          + address.getHostName() + "].";
      if (reconnectionDelay > 0) {
        msg += " We will try again later.";
        fireConnector(); // fire the connector thread
      }
      LogLog.error(msg, e);
    }
  }

  private void fireConnector()
  {
    if (connector == null) {
      LogLog.debug("Starting a new connector thread.");
      connector = new Connector();
      connector.setDaemon(true);
      connector.setPriority(Thread.MIN_PRIORITY);
      connector.start();
    }
  }

  private InetAddress getAddressByName(String host)
  {
    try {
      return InetAddress.getByName(host);
    } catch (Exception e) {
      LogLog.error("Could not find address of [" + host + "].", e);
      return null;
    }
  }

  /**
   * The Connector will reconnect when the server becomes available again. It
   * does this by attempting to open a new connection every
   * <code>reconnectionDelay</code> milliseconds.
   * <p>
   * It stops trying whenever a connection is established. It will restart to
   * try reconnect to the server when previpously open connection is droppped.
   * 
   * @author Ceki G&uuml;lc&uuml;
   * @since 0.8.4
   */
  private class Connector extends Thread
  {
    boolean interrupted = false;

    public void run()
    {
      Socket socket;
      while (!interrupted) {
        try {
          sleep(reconnectionDelay);
          LogLog.debug("Attempting connection to " + address.getHostName());
          socket = new Socket(address, port);
          synchronized (this) {
            oos = new ObjectOutputStream(socket.getOutputStream());
            connector = null;
            LogLog.debug("Connection established. Exiting connector thread.");
            break;
          }
        } catch (InterruptedException e) {
          LogLog.debug("Connector interrupted. Leaving loop.");
          return;
        } catch (java.net.ConnectException e) {
          LogLog.debug("Remote host " + address.getHostName()
              + " refused connection.");
        } catch (IOException e) {
          LogLog.debug("Could not connect to " + address.getHostName()
              + ". Exception is " + e);
        }
      }
    }
  }
}
